// This is free and unencumbered software released into the public
// domain.  For more information, see <http://unlicense.org> or the
// accompanying UNLICENSE file.

package mock_test

import (
	"errors"
	"fmt"
	"go/ast"
	"go/format"
	"testing"

	"git.sr.ht/~nelsam/hel/mock"
	"github.com/poy/onpar/expect"
)

func TestMockSimpleMethod(t *testing.T) {
	expect := expect.New(t)

	spec := typeSpec(expect, `
 type Foo interface {
         Foo()
 }`)
	m, err := mock.For(spec.Name.Name, spec.TypeParams.List, spec.Type.(*ast.InterfaceType))
	expect(err).To(not(haveOccurred()))
	method := mock.MethodFor(m, "Foo", method(expect, spec))

	expected, err := format.Source([]byte(`
 package foo

 func (m *mockFoo) Foo() {
   m.t.Helper()
   m.FooCalled <- true
 }`))
	expect(err).To(not(haveOccurred()))

	src := source(expect, "foo", []ast.Decl{method.Ast()}, nil)
	expect(src).To(equal(string(expected)))

	fields := method.Fields()
	expect(fields).To(haveLen(1))

	expect(fields[0].Names[0].Name).To(equal("FooCalled"))
	ch, ok := fields[0].Type.(*ast.ChanType)
	expect(ok).To(beTrue())
	expect(ch.Dir).To(equal(ast.SEND | ast.RECV))
	ident, ok := ch.Value.(*ast.Ident)
	expect(ident.Name).To(equal("bool"))
}

func TestMockMethodParams(t *testing.T) {
	expect := expect.New(t)

	spec := typeSpec(expect, `
 type Foo interface {
         Foo(foo, bar string, baz int)
 }`)
	m, err := mock.For(spec.Name.Name, spec.TypeParams.List, spec.Type.(*ast.InterfaceType))
	expect(err).To(not(haveOccurred()))
	method := mock.MethodFor(m, "Foo", method(expect, spec))

	expected, err := format.Source([]byte(`
 package foo

 func (m *mockFoo) Foo(foo, bar string, baz int) {
   m.t.Helper()
   m.FooCalled <- true
   m.FooInput.Foo <- foo
   m.FooInput.Bar <- bar
   m.FooInput.Baz <- baz
 }`))
	expect(err).To(not(haveOccurred()))

	src := source(expect, "foo", []ast.Decl{method.Ast()}, nil)
	expect(src).To(equal(string(expected)))

	fields := method.Fields()
	expect(fields).To(haveLen(2))

	expect(fields[0].Names[0].Name).To(equal("FooCalled"))
	ch, ok := fields[0].Type.(*ast.ChanType)
	expect(ok).To(beTrue())
	expect(ch.Dir).To(equal(ast.SEND | ast.RECV))
	ident, ok := ch.Value.(*ast.Ident)
	expect(ident.Name).To(equal("bool"))

	expect(fields[1].Names[0].Name).To(equal("FooInput"))
	input, ok := fields[1].Type.(*ast.StructType)
	expect(ok).To(beTrue())
	expect(input.Fields.List).To(haveLen(2))

	fooBar := input.Fields.List[0]
	expect(fooBar.Names).To(haveLen(2))
	expect(fooBar.Names[0].Name).To(equal("Foo"))
	expect(fooBar.Names[1].Name).To(equal("Bar"))
	ch, ok = fooBar.Type.(*ast.ChanType)
	expect(ok).To(beTrue())
	expect(ch.Dir).To(equal(ast.SEND | ast.RECV))
	ident, ok = ch.Value.(*ast.Ident)
	expect(ident.Name).To(equal("string"))

	baz := input.Fields.List[1]
	expect(baz.Names[0].Name).To(equal("Baz"))
	ch, ok = baz.Type.(*ast.ChanType)
	expect(ok).To(beTrue())
	expect(ch.Dir).To(equal(ast.SEND | ast.RECV))
	ident, ok = ch.Value.(*ast.Ident)
	expect(ident.Name).To(equal("int"))
}

func TestMockMethodReturns(t *testing.T) {
	expect := expect.New(t)

	spec := typeSpec(expect, `
 type Foo interface {
   Foo() (foo, bar string, baz int)
 }`)
	m, err := mock.For(spec.Name.Name, spec.TypeParams.List, spec.Type.(*ast.InterfaceType))
	expect(err).To(not(haveOccurred()))
	method := mock.MethodFor(m, "Foo", method(expect, spec))

	expected, err := format.Source([]byte(`
 package foo

 func (m *mockFoo) Foo() (foo, bar string, baz int) {
   m.t.Helper()
   m.FooCalled <- true
   vegr.PopulateReturns(m.t, "Foo", m.timeout, m.FooOutput, &foo, &bar, &baz)
   return foo, bar, baz
 }`))
	expect(err).To(not(haveOccurred()))

	src := source(expect, "foo", []ast.Decl{method.Ast()}, nil)
	expect(src).To(equal(string(expected)))

	fields := method.Fields()
	expect(fields).To(haveLen(2))

	expect(fields[0].Names[0].Name).To(equal("FooCalled"))
	ch, ok := fields[0].Type.(*ast.ChanType)
	expect(ok).To(beTrue())
	expect(ch.Dir).To(equal(ast.SEND | ast.RECV))
	ident, ok := ch.Value.(*ast.Ident)
	expect(ident.Name).To(equal("bool"))

	expect(fields[1].Names[0].Name).To(equal("FooOutput"))
	input, ok := fields[1].Type.(*ast.StructType)
	expect(ok).To(beTrue())
	expect(input.Fields.List).To(haveLen(2))

	fooBar := input.Fields.List[0]
	expect(fooBar.Names).To(haveLen(2))
	expect(fooBar.Names[0].Name).To(equal("Foo"))
	expect(fooBar.Names[1].Name).To(equal("Bar"))
	ch, ok = fooBar.Type.(*ast.ChanType)
	expect(ok).To(beTrue())
	expect(ch.Dir).To(equal(ast.SEND | ast.RECV))
	ident, ok = ch.Value.(*ast.Ident)
	expect(ident.Name).To(equal("string"))

	baz := input.Fields.List[1]
	expect(baz.Names[0].Name).To(equal("Baz"))
	ch, ok = baz.Type.(*ast.ChanType)
	expect(ok).To(beTrue())
	expect(ch.Dir).To(equal(ast.SEND | ast.RECV))
	ident, ok = ch.Value.(*ast.Ident)
	expect(ident.Name).To(equal("int"))
}

func TestMockMethodUnnamedValues(t *testing.T) {
	expect := expect.New(t)

	spec := typeSpec(expect, `
 type Foo interface {
   Foo(int, string) (string, error)
 }`)
	m, err := mock.For(spec.Name.Name, spec.TypeParams.List, spec.Type.(*ast.InterfaceType))
	expect(err).To(not(haveOccurred()))
	method := mock.MethodFor(m, "Foo", method(expect, spec))

	expected, err := format.Source([]byte(`
 package foo

 func (m *mockFoo) Foo(arg0 int, arg1 string) (ret0 string, ret1 error) {
   m.t.Helper()
   m.FooCalled <- true
   m.FooInput.Arg0 <- arg0
   m.FooInput.Arg1 <- arg1
   vegr.PopulateReturns(m.t, "Foo", m.timeout, m.FooOutput, &ret0, &ret1)
   return ret0, ret1
 }`))
	expect(err).To(not(haveOccurred()))

	src := source(expect, "foo", []ast.Decl{method.Ast()}, nil)
	expect(src).To(equal(string(expected)))
}

func TestMockMethodLocalTypes(t *testing.T) {
	expect := expect.New(t)

	spec := typeSpec(expect, `
 type Foo interface {
   Foo(bar bar.Bar, baz func(f Foo) error, eggs ...Egg) (*Foo, func() Foo, error)
 }`)
	m, err := mock.For(spec.Name.Name, spec.TypeParams.List, spec.Type.(*ast.InterfaceType))
	expect(err).To(not(haveOccurred()))
	method := mock.MethodFor(m, "Foo", method(expect, spec))

	expected, err := format.Source([]byte(`
 package foo

 func (m *mockFoo) Foo(bar bar.Bar, baz func(f Foo) error, eggs ...Egg) (ret0 *Foo, ret1 func() Foo, ret2 error) {
   m.t.Helper()
   m.FooCalled <- true
   m.FooInput.Bar <- bar
   m.FooInput.Baz <- baz
   m.FooInput.Eggs <- eggs
   vegr.PopulateReturns(m.t, "Foo", m.timeout,  m.FooOutput, &ret0, &ret1, &ret2)
   return ret0, ret1, ret2
 }`))
	expect(err).To(not(haveOccurred()))

	src := source(expect, "foo", []ast.Decl{method.Ast()}, nil)
	expect(src).To(equal(string(expected)))

	err = method.PrependLocalPackage("foo")
	expect(err).To(not(haveOccurred()))

	expected, err = format.Source([]byte(`
 package foo

 func (m *mockFoo) Foo(bar bar.Bar, baz func(f foo.Foo) error, eggs ...foo.Egg) (ret0 *foo.Foo, ret1 func() foo.Foo, ret2 error) {
   m.t.Helper()
   m.FooCalled <- true
   m.FooInput.Bar <- bar
   m.FooInput.Baz <- baz
   m.FooInput.Eggs <- eggs
   vegr.PopulateReturns(m.t, "Foo", m.timeout,  m.FooOutput, &ret0, &ret1, &ret2)
   return ret0, ret1, ret2
 }`))
	expect(err).To(not(haveOccurred()))

	src = source(expect, "foo", []ast.Decl{method.Ast()}, nil)
	expect(src).To(equal(string(expected)))
}

func TestMockPrependLocalPackageBuiltinTypes(t *testing.T) {
	for _, tt := range []string{
		"bool",
		"string",
		"int", "int8", "int16", "int32", "int64",
		"uint", "uint8", "uint16", "uint32", "uint64", "uintptr",
		"byte",
		"rune",
		"float32", "float64",
		"complex64", "complex128",
		"error",
		"any", "interface{}",
	} {
		tt := tt
		t.Run("param type "+tt, func(t *testing.T) {
			t.Parallel()

			expect := expect.New(t)

			spec := typeSpec(expect, fmt.Sprintf(`
 type Foo interface {
   Foo(%s)
 }`, tt))
			m, err := mock.For(spec.Name.Name, spec.TypeParams.List, spec.Type.(*ast.InterfaceType))
			expect(err).To(not(haveOccurred()))
			method := mock.MethodFor(m, "Foo", method(expect, spec))

			err = method.PrependLocalPackage("foo")
			expect(err).To(not(haveOccurred()))
		})

		t.Run("result type "+tt, func(t *testing.T) {
			t.Parallel()

			expect := expect.New(t)

			spec := typeSpec(expect, fmt.Sprintf(`
 type Foo interface {
   Foo() %s
 }`, tt))
			m, err := mock.For(spec.Name.Name, spec.TypeParams.List, spec.Type.(*ast.InterfaceType))
			expect(err).To(not(haveOccurred()))
			method := mock.MethodFor(m, "Foo", method(expect, spec))

			err = method.PrependLocalPackage("foo")
			expect(err).To(not(haveOccurred()))
		})
	}
}

func TestMockMethodPrependPackageUnexportedParams(t *testing.T) {
	expect := expect.New(t)

	spec := typeSpec(expect, `
 type Foo interface {
   Foo(bar bar.Bar, baz func(f foo) error, eggs ...Egg) (*Foo, func() Foo, error)
 }`)
	m, err := mock.For(spec.Name.Name, spec.TypeParams.List, spec.Type.(*ast.InterfaceType))
	expect(err).To(not(haveOccurred()))
	method := mock.MethodFor(m, "Foo", method(expect, spec))

	err = method.PrependLocalPackage("foo")
	expect(errors.Is(err, mock.ErrUnexported)).To(equal(true))
}

func TestMockMethodPrependPackageUnexportedResults(t *testing.T) {
	expect := expect.New(t)

	spec := typeSpec(expect, `
 type Foo interface {
   Foo(bar bar.Bar, baz func(f Foo) error, eggs ...Egg) (*Foo, func() foo, error)
 }`)
	m, err := mock.For(spec.Name.Name, spec.TypeParams.List, spec.Type.(*ast.InterfaceType))
	expect(err).To(not(haveOccurred()))
	method := mock.MethodFor(m, "Foo", method(expect, spec))

	err = method.PrependLocalPackage("foo")
	expect(errors.Is(err, mock.ErrUnexported)).To(equal(true))
}

func TestMockMethodLocalTypeNesting(t *testing.T) {
	expect := expect.New(t)

	spec := typeSpec(expect, `
 type Foo interface {
   Foo(bar []Bar, bacon map[Foo]Bar) (baz []Baz, eggs map[Foo]Bar)
 }`)
	m, err := mock.For(spec.Name.Name, spec.TypeParams.List, spec.Type.(*ast.InterfaceType))
	expect(err).To(not(haveOccurred()))
	method := mock.MethodFor(m, "Foo", method(expect, spec))
	err = method.PrependLocalPackage("foo")
	expect(err).To(not(haveOccurred()))

	expected, err := format.Source([]byte(`
 package foo

 func (m *mockFoo) Foo(bar []foo.Bar, bacon map[foo.Foo]foo.Bar) (baz []foo.Baz, eggs map[foo.Foo]foo.Bar) {
   m.t.Helper()
   m.FooCalled <- true
   m.FooInput.Bar <- bar
   m.FooInput.Bacon <- bacon
   vegr.PopulateReturns(m.t, "Foo", m.timeout, m.FooOutput, &baz, &eggs)
   return baz, eggs
 }`))
	expect(err).To(not(haveOccurred()))

	src := source(expect, "foo", []ast.Decl{method.Ast()}, nil)
	expect(src).To(equal(string(expected)))
}

func TestMockMethodChannelTypeNesting(t *testing.T) {
	expect := expect.New(t)

	spec := typeSpec(expect, `
 type Foo interface {
   Foo(bar chan Bar) (chan Bar)
 }`)
	m, err := mock.For(spec.Name.Name, spec.TypeParams.List, spec.Type.(*ast.InterfaceType))
	expect(err).To(not(haveOccurred()))
	method := mock.MethodFor(m, "Foo", method(expect, spec))
	err = method.PrependLocalPackage("foo")
	expect(err).To(not(haveOccurred()))

	expected, err := format.Source([]byte(`
 package foo

 func (m *mockFoo) Foo(bar chan foo.Bar) (ret0 chan foo.Bar) {
   m.t.Helper()
   m.FooCalled <- true
   m.FooInput.Bar <- bar
   vegr.PopulateReturns(m.t, "Foo", m.timeout, m.FooOutput, &ret0)
   return ret0
 }`))
	expect(err).To(not(haveOccurred()))

	src := source(expect, "foo", []ast.Decl{method.Ast()}, nil)
	expect(src).To(equal(string(expected)))
}

func TestMockMethodUnderscoreNames(t *testing.T) {
	expect := expect.New(t)

	spec := typeSpec(expect, `
 type Foo interface {
         Foo(_ string)
 }`)
	m, err := mock.For(spec.Name.Name, spec.TypeParams.List, spec.Type.(*ast.InterfaceType))
	expect(err).To(not(haveOccurred()))
	method := mock.MethodFor(m, "Foo", method(expect, spec))

	expected, err := format.Source([]byte(`
 package foo

 func (m *mockFoo) Foo(arg0 string) {
   m.t.Helper()
   m.FooCalled <- true
   m.FooInput.Arg0 <- arg0
 }`))
	expect(err).To(not(haveOccurred()))

	src := source(expect, "foo", []ast.Decl{method.Ast()}, nil)
	expect(src).To(equal(string(expected)))
}

func TestMockMethodReceiverNameConflicts(t *testing.T) {
	expect := expect.New(t)

	spec := typeSpec(expect, `
 type Foo interface {
         Foo(m string)
         Bar() (m string)
 }`)
	m, err := mock.For(spec.Name.Name, spec.TypeParams.List, spec.Type.(*ast.InterfaceType))
	expect(err).To(not(haveOccurred()))

	t.Run("args", func(t *testing.T) {
		method := mock.MethodFor(m, "Foo", method(expect, spec, "Foo"))

		expected, err := format.Source([]byte(`
 package foo

 func (m *mockFoo) Foo(m_ string) {
   m.t.Helper()
   m.FooCalled <- true
   m.FooInput.M <- m_
 }`))
		expect(err).To(not(haveOccurred()))

		src := source(expect, "foo", []ast.Decl{method.Ast()}, nil)
		expect(src).To(equal(string(expected)))
	})

	t.Run("results", func(t *testing.T) {
		method := mock.MethodFor(m, "Bar", method(expect, spec, "Bar"))

		expected, err := format.Source([]byte(`
 package foo

 func (m *mockFoo) Bar() (m_ string) {
   m.t.Helper()
   m.BarCalled <- true
   vegr.PopulateReturns(m.t, "Bar", m.timeout, m.BarOutput, &m_)
   return m_
 }`))
		expect(err).To(not(haveOccurred()))

		src := source(expect, "foo", []ast.Decl{method.Ast()}, nil)
		expect(src).To(equal(string(expected)))
	})
}
