// This is free and unencumbered software released into the public
// domain.  For more information, see <http://unlicense.org> or the
// accompanying UNLICENSE file.

package mock_test

import (
	"bytes"
	"go/ast"
	"go/format"
	"testing"
	"time"

	"git.sr.ht/~nelsam/hel/dep"
	"git.sr.ht/~nelsam/hel/mock"
	"git.sr.ht/~nelsam/hel/pkg/pers"
	"git.sr.ht/~nelsam/hel/typ"
	"github.com/fatih/color"
	"github.com/poy/onpar"
	"github.com/poy/onpar/diff"
	"github.com/poy/onpar/expect"
)

func TestMocks(t *testing.T) {
	type testCtx struct {
		t      *testing.T
		expect expect.Expectation
		finder *mockTypeFinder
	}

	o := onpar.BeforeEach(onpar.New(t), func(t *testing.T) testCtx {
		differ := diff.New(
			diff.Actual(diff.WithSprinter(color.New(color.BgRed, color.FgBlack))),
			diff.Expected(diff.WithSprinter(color.New(color.BgYellow, color.FgBlack))),
		)
		return testCtx{t: t, expect: expect.New(t, expect.WithDiffer(differ)), finder: newMockTypeFinder(t, time.Second)}
	})
	defer o.Run()

	o.Group("without local interfaces", func() {
		o.Spec("it skips types which are not directly depended upon by concrete logic", func(tt testCtx) {
			decls := []ast.Decl{
				&ast.GenDecl{Specs: []ast.Spec{typeSpec(tt.expect, `
  type Foo interface {
   Bar() int
  }`)}},
				&ast.GenDecl{Specs: []ast.Spec{typeSpec(tt.expect, `
  type Bar interface {
   Foo(foo string)
   Baz()
  }`)}},
			}

			pers.ConsistentlyReturn(tt.t, tt.finder.DependenciesOutput, nil)
			pers.Return(tt.finder.DeclsOutput, decls)
			m, err := mock.Generate(tt.finder, mock.ForConcreteDependencies())
			tt.expect(err).To(not(haveOccurred()))
			tt.expect(m.Len()).To(equal(0))
		})

		o.Spec("it looks for dependencies of struct types", func(tt testCtx) {
			depSpec := typeSpec(tt.expect, `
  type Foo interface {
   Bar() int
  }`)
			decls := []ast.Decl{
				&ast.GenDecl{Specs: []ast.Spec{depSpec}},
				&ast.GenDecl{Specs: []ast.Spec{typeSpec(tt.expect, `
  type Bar struct {
   Fooer Foo
  }`)}},
			}

			pers.Return(tt.finder.DeclsOutput, decls)
			var (
				m    *mock.Mocks
				err  error
				done = make(chan struct{})
			)
			go func() {
				m, err = mock.Generate(tt.finder, mock.ForConcreteDependencies())
				close(done)
			}()
			tt.expect(tt.finder).To(haveMethodExecuted(
				"Dependencies",
				within(time.Second),
				withArgs("Bar"),
				returning([]typ.Dependency{dep.New(depSpec.Name, depSpec.Type)}),
			))
			<-done
			tt.expect(err).To(not(haveOccurred()))
			tt.expect(m.Len()).To(equal(1))
			tt.expect(m.Index(0)).To(equal(mockFor(tt.expect, decls[0])))
		})

		o.Spec("it looks for dependencies of function types", func(tt testCtx) {
			depSpec := typeSpec(tt.expect, `
  type Foo interface {
   Bar() int
  }`)
			decls := []ast.Decl{
				&ast.GenDecl{Specs: []ast.Spec{depSpec}},
				function(tt.expect, `
  func Bar(f Foo) int {
   return -1
  }`),
			}

			pers.Return(tt.finder.DeclsOutput, decls)
			var (
				m    *mock.Mocks
				err  error
				done = make(chan struct{})
			)
			go func() {
				m, err = mock.Generate(tt.finder, mock.ForConcreteDependencies())
				close(done)
			}()
			tt.expect(tt.finder).To(haveMethodExecuted(
				"Dependencies",
				within(time.Second),
				withArgs("Bar"),
				returning([]typ.Dependency{dep.New(depSpec.Name, depSpec.Type)}),
			))
			<-done
			tt.expect(err).To(not(haveOccurred()))
			tt.expect(m.Len()).To(equal(1))
			tt.expect(m.Index(0)).To(equal(mockFor(tt.expect, decls[0])))
		})
	})

	o.Spec("it finds exported interface types to mock", func(tt testCtx) {
		decls := []ast.Decl{
			&ast.GenDecl{Specs: []ast.Spec{typeSpec(tt.expect, `
  type Foo interface {
   Bar() int
  }`)}},
			&ast.GenDecl{Specs: []ast.Spec{typeSpec(tt.expect, `
  type Bar interface {
   Foo(foo string)
   Baz()
  }`)}},
		}

		pers.ConsistentlyReturn(tt.t, tt.finder.DependenciesOutput, nil)
		pers.Return(tt.finder.DeclsOutput, decls)
		m, err := mock.Generate(tt.finder)
		tt.expect(err).To(not(haveOccurred()))
		tt.expect(m.Len()).To(equal(2))
		tt.expect(m.Index(0)).To(equal(mockFor(tt.expect, decls[0])))
		tt.expect(m.Index(1)).To(equal(mockFor(tt.expect, decls[1])))
	})

	o.Spec("it generates go output from mocks", func(tt testCtx) {
		decls := []ast.Decl{
			&ast.GenDecl{Specs: []ast.Spec{typeSpec(tt.expect, `
  type Foo interface {
   Bar() int
  }`)}},
			&ast.GenDecl{Specs: []ast.Spec{typeSpec(tt.expect, `
  type Bar interface {
   Foo(foo string) Foo
   Baz()
   Bacon(func(Eggs) Eggs) func(Eggs) Eggs
  }`)}},
		}

		pers.ConsistentlyReturn(tt.t, tt.finder.DependenciesOutput, nil)
		pers.Return(tt.finder.DeclsOutput, decls)

		m, err := mock.Generate(tt.finder)
		tt.expect(err).To(not(haveOccurred()))

		var buf bytes.Buffer
		err = m.Output("foo", "test/withoutimports", 100, &buf)
		tt.expect(err).To(not(haveOccurred()))

		expected, err := format.Source([]byte(`
 // Code generated by git.sr.ht/~nelsam/hel. DO NOT EDIT.
 //
 // This file contains mocks generated by hel. Do not edit this code by
 // hand unless you *really* know what you're doing. Expect any changes
 // made manually to be overwritten the next time hel regenerates this
 // file.

 package foo

 import (
  "git.sr.ht/~nelsam/hel/vegr"
  "time"
 )

 type mockFoo struct {
  t vegr.T
  timeout time.Duration
  BarCalled chan bool
  BarOutput struct {
   Ret0 chan int
  }
 }

 func newMockFoo(t vegr.T, timeout time.Duration) *mockFoo {
  m := &mockFoo{t: t, timeout: timeout}
  m.BarCalled = make(chan bool, 100)
  m.BarOutput.Ret0 = make(chan int, 100)
  return m
 }
 func (m *mockFoo) Bar() (ret0 int) {
  m.t.Helper()
  m.BarCalled <- true
  vegr.PopulateReturns(m.t, "Bar", m.timeout, m.BarOutput, &ret0)
  return ret0
 }

 type mockBar struct {
  t vegr.T
  timeout time.Duration
  FooCalled chan bool
  FooInput struct {
   Foo chan string
  }
  FooOutput struct {
   Ret0 chan Foo
  }
  BazCalled chan bool
  BaconCalled chan bool
  BaconInput struct {
    Arg0 chan func(Eggs) Eggs
  }
  BaconOutput struct {
    Ret0 chan func(Eggs) Eggs
  }
 }

 func newMockBar(t vegr.T, timeout time.Duration) *mockBar {
  m := &mockBar{t: t, timeout: timeout}
  m.FooCalled = make(chan bool, 100)
  m.FooInput.Foo = make(chan string, 100)
  m.FooOutput.Ret0 = make(chan Foo, 100)
  m.BazCalled = make(chan bool, 100)
  m.BaconCalled = make(chan bool, 100)
  m.BaconInput.Arg0 = make(chan func(Eggs) Eggs, 100)
  m.BaconOutput.Ret0 = make(chan func(Eggs) Eggs, 100)
  return m
 }
 func (m *mockBar) Foo(foo string) (ret0 Foo) {
  m.t.Helper()
  m.FooCalled <- true
  m.FooInput.Foo <- foo
  vegr.PopulateReturns(m.t, "Foo", m.timeout, m.FooOutput, &ret0)
  return ret0
 }
 func (m *mockBar) Baz() {
  m.t.Helper()
  m.BazCalled <- true
 }
 func (m *mockBar) Bacon(arg0 func(Eggs) Eggs) (ret0 func(Eggs) Eggs) {
  m.t.Helper()
  m.BaconCalled <- true
  m.BaconInput.Arg0 <- arg0
  vegr.PopulateReturns(m.t, "Bacon", m.timeout, m.BaconOutput, &ret0)
  return ret0
 }
 `))
		tt.expect(err).To(not(haveOccurred()))
		tt.expect(buf.String()).To(equal(string(expected)))
	})

	o.Spec("it generates go output with the local package selector added to local types", func(tt testCtx) {
		decls := []ast.Decl{
			&ast.GenDecl{Specs: []ast.Spec{typeSpec(tt.expect, `
  type Foo interface {
   Bar() int
  }`)}},
			&ast.GenDecl{Specs: []ast.Spec{typeSpec(tt.expect, `
  type Bar interface {
   Foo(foo string) Foo
   Baz()
   Bacon(func(Eggs) Eggs) func(Eggs) Eggs
  }`)}},
		}

		pers.ConsistentlyReturn(tt.t, tt.finder.DependenciesOutput, nil)
		pers.Return(tt.finder.DeclsOutput, decls)

		m, err := mock.Generate(tt.finder)
		tt.expect(err).To(not(haveOccurred()))

		m.PrependLocalPackage("foo")
		var buf bytes.Buffer
		m.Output("foo_test", "test/withoutimports", 100, &buf)

		expected, err := format.Source([]byte(`
 // Code generated by git.sr.ht/~nelsam/hel. DO NOT EDIT.
 //
 // This file contains mocks generated by hel. Do not edit this code by
 // hand unless you *really* know what you're doing. Expect any changes
 // made manually to be overwritten the next time hel regenerates this
 // file.

 package foo_test

 import (
  "git.sr.ht/~nelsam/hel/vegr"
  "time"
 )

 type mockFoo struct {
  t vegr.T
  timeout time.Duration
  BarCalled chan bool
  BarOutput struct {
   Ret0 chan int
  }
 }

 func newMockFoo(t vegr.T, timeout time.Duration) *mockFoo {
  m := &mockFoo{t: t, timeout: timeout}
  m.BarCalled = make(chan bool, 100)
  m.BarOutput.Ret0 = make(chan int, 100)
  return m
 }
 func (m *mockFoo) Bar() (ret0 int) {
  m.t.Helper()
  m.BarCalled <- true
  vegr.PopulateReturns(m.t, "Bar", m.timeout, m.BarOutput, &ret0)
  return ret0
 }

 type mockBar struct {
  t vegr.T
  timeout time.Duration
  FooCalled chan bool
  FooInput struct {
   Foo chan string
  }
  FooOutput struct {
   Ret0 chan foo.Foo
  }
  BazCalled chan bool
  BaconCalled chan bool
  BaconInput struct {
    Arg0 chan func(foo.Eggs) foo.Eggs
  }
  BaconOutput struct {
    Ret0 chan func(foo.Eggs) foo.Eggs
  }
 }

 func newMockBar(t vegr.T, timeout time.Duration) *mockBar {
  m := &mockBar{t: t, timeout: timeout}
  m.FooCalled = make(chan bool, 100)
  m.FooInput.Foo = make(chan string, 100)
  m.FooOutput.Ret0 = make(chan foo.Foo, 100)
  m.BazCalled = make(chan bool, 100)
  m.BaconCalled = make(chan bool, 100)
  m.BaconInput.Arg0 = make(chan func(foo.Eggs) foo.Eggs, 100)
  m.BaconOutput.Ret0 = make(chan func(foo.Eggs) foo.Eggs, 100)
  return m
 }
 func (m *mockBar) Foo(foo string) (ret0 foo.Foo) {
  m.t.Helper()
  m.FooCalled <- true
  m.FooInput.Foo <- foo
  vegr.PopulateReturns(m.t, "Foo", m.timeout, m.FooOutput, &ret0)
  return ret0
 }
 func (m *mockBar) Baz() {
  m.t.Helper()
  m.BazCalled <- true
 }
 func (m *mockBar) Bacon(arg0 func(foo.Eggs) foo.Eggs) (ret0 func(foo.Eggs) foo.Eggs) {
  m.t.Helper()
  m.BaconCalled <- true
  m.BaconInput.Arg0 <- arg0
  vegr.PopulateReturns(m.t, "Bacon", m.timeout, m.BaconOutput, &ret0)
  return ret0
 }
 `))
		tt.expect(err).To(not(haveOccurred()))
		tt.expect(buf.String()).To(equal(string(expected)))
	})

	o.Spec("it outputs mocks for interface types that are dependencies of generated mocks", func(tt testCtx) {
		decls := []ast.Decl{
			&ast.GenDecl{Specs: []ast.Spec{typeSpec(tt.expect, `
  type Bar interface {
    Foo(foo string) (Foo, b.Foo)
  }`)}},
			&ast.GenDecl{Specs: []ast.Spec{typeSpec(tt.expect, `
  type Foo interface {
    Foo() string
  }`)}},
		}

		barDeps := []typ.Dependency{
			// We shouldn't see duplicates of types we're
			// already mocking.
			depSpec(tt.expect, `
			type Foo interface{
				Foo() string
			}`),
			// Different package names should have the type
			// name altered.
			depSpec(tt.expect, `
			type Foo interface{
				Foo() string
			}`, dep.FromPkg("b", "b")),
			// Different types entirely should be supported.
			depSpec(tt.expect, `
			type Baz interface {
				Baz() Baz
			}
			`, dep.FromPkg("foo", "some/path/to/foo"), dep.WithImportAlias("baz")),
		}
		fooDeps := []typ.Dependency{
			// We shouldn't see duplicate dependencies from
			// previous types, either.
			depSpec(tt.expect, `
			type Baz interface {
				Baz() Baz
			}
			`, dep.FromPkg("foo", "some/path/to/foo")),
		}

		pers.Return(tt.finder.DeclsOutput, decls)
		pers.Return(tt.finder.DependenciesOutput, barDeps)
		pers.Return(tt.finder.DependenciesOutput, fooDeps)
		m, err := mock.Generate(tt.finder)
		tt.expect(err).To(not(haveOccurred()))

		buf := bytes.Buffer{}
		err = m.Output("foo", "test/withoutimports", 100, &buf)
		tt.expect(err).To(not(haveOccurred()))

		// TODO: For some reason, functions are coming out without
		// whitespace between them. We need to figure that out.
		expected, err := format.Source([]byte(`
 // Code generated by git.sr.ht/~nelsam/hel. DO NOT EDIT.
 //
 // This file contains mocks generated by hel. Do not edit this code by
 // hand unless you *really* know what you're doing. Expect any changes
 // made manually to be overwritten the next time hel regenerates this
 // file.

 package foo

 import (
  "b"
  "git.sr.ht/~nelsam/hel/vegr"
  baz "some/path/to/foo"
  "time"
 )

 type mockBar struct {
  t vegr.T
  timeout time.Duration
  FooCalled chan bool
  FooInput struct {
    Foo chan string
  }
  FooOutput struct {
    Ret0 chan Foo
    Ret1 chan b.Foo
  }
 }

 func newMockBar(t vegr.T, timeout time.Duration) *mockBar {
  m := &mockBar{t: t, timeout: timeout}
  m.FooCalled = make(chan bool, 100)
  m.FooInput.Foo = make(chan string, 100)
  m.FooOutput.Ret0 = make(chan Foo, 100)
  m.FooOutput.Ret1 = make(chan b.Foo, 100)
  return m
 }
 func (m *mockBar) Foo(foo string) (ret0 Foo, ret1 b.Foo) {
  m.t.Helper()
  m.FooCalled <- true
  m.FooInput.Foo <- foo
  vegr.PopulateReturns(m.t, "Foo", m.timeout, m.FooOutput, &ret0, &ret1)
  return ret0, ret1
 }

 type mockFoo struct {
  t vegr.T
  timeout time.Duration
  FooCalled chan bool
  FooOutput struct {
   Ret0 chan string
  }
 }

 func newMockFoo(t vegr.T, timeout time.Duration) *mockFoo {
  m := &mockFoo{t: t, timeout: timeout}
  m.FooCalled = make(chan bool, 100)
  m.FooOutput.Ret0 = make(chan string, 100)
  return m
 }
 func (m *mockFoo) Foo() (ret0 string) {
  m.t.Helper()
  m.FooCalled <- true
  vegr.PopulateReturns(m.t, "Foo", m.timeout, m.FooOutput, &ret0)
  return ret0
 }

 type mockBFoo struct {
  t vegr.T
  timeout time.Duration
  FooCalled chan bool
  FooOutput struct {
   Ret0 chan string
  }
 }

 func newMockBFoo(t vegr.T, timeout time.Duration) *mockBFoo {
  m := &mockBFoo{t: t, timeout: timeout}
  m.FooCalled = make(chan bool, 100)
  m.FooOutput.Ret0 = make(chan string, 100)
  return m
 }
 func (m *mockBFoo) Foo() (ret0 string) {
  m.t.Helper()
  m.FooCalled <- true
  vegr.PopulateReturns(m.t, "Foo", m.timeout, m.FooOutput, &ret0)
  return ret0
 }

 type mockBaz struct {
  t vegr.T
  timeout time.Duration
  BazCalled chan bool
  BazOutput struct {
   Ret0 chan baz.Baz
  }
 }

 func newMockBaz(t vegr.T, timeout time.Duration) *mockBaz {
  m := &mockBaz{t: t, timeout: timeout}
  m.BazCalled = make(chan bool, 100)
  m.BazOutput.Ret0 = make(chan baz.Baz, 100)
  return m
 }
 func (m *mockBaz) Baz() (ret0 baz.Baz) {
  m.t.Helper()
  m.BazCalled <- true
  vegr.PopulateReturns(m.t, "Baz", m.timeout, m.BazOutput, &ret0)
  return ret0
 }
 `))
		tt.expect(err).To(not(haveOccurred()))
		tt.expect(buf.String()).To(equal(string(expected)))
	})

	o.Spec("it can handle variable names shadowing package names", func(tt testCtx) {
		decls := []ast.Decl{
			&ast.GenDecl{Specs: []ast.Spec{typeSpec(tt.expect, `
  type Foo interface {
   Bar() int
  }`)}},
			&ast.GenDecl{Specs: []ast.Spec{typeSpec(tt.expect, `
  type Bar interface {
   Foo(bar bar.Bar) Foo
   Baz()
   Bacon(func(Eggs) Eggs) func(Eggs) Eggs
  }`)}},
		}

		pers.Return(tt.finder.DependenciesOutput, []typ.Dependency{
			depSpec(tt.expect, `
  type Bar string`, dep.FromPkg("bar", "a/path/to/bar")),
		})
		pers.ConsistentlyReturn(tt.t, tt.finder.DependenciesOutput, nil)
		pers.Return(tt.finder.DeclsOutput, decls)
		m, err := mock.Generate(tt.finder)
		tt.expect(err).To(not(haveOccurred()))

		var buf bytes.Buffer
		err = m.Output("foo", "test/withimports", 100, &buf)
		tt.expect(err).To(not(haveOccurred()))

		expected, err := format.Source([]byte(`
 // Code generated by git.sr.ht/~nelsam/hel. DO NOT EDIT.
 //
 // This file contains mocks generated by hel. Do not edit this code by
 // hand unless you *really* know what you're doing. Expect any changes
 // made manually to be overwritten the next time hel regenerates this
 // file.

 package foo

 import (
    "a/path/to/bar"
    "git.sr.ht/~nelsam/hel/vegr"
    "time"
 )

 type mockFoo struct {
  t vegr.T
  timeout time.Duration
  BarCalled chan bool
  BarOutput struct {
   Ret0 chan int
  }
 }

 func newMockFoo(t vegr.T, timeout time.Duration) *mockFoo {
  m := &mockFoo{t: t, timeout: timeout}
  m.BarCalled = make(chan bool, 100)
  m.BarOutput.Ret0 = make(chan int, 100)
  return m
 }
 func (m *mockFoo) Bar() (ret0 int) {
  m.t.Helper()
  m.BarCalled <- true
  vegr.PopulateReturns(m.t, "Bar", m.timeout, m.BarOutput, &ret0)
  return ret0
 }

 type mockBar struct {
  t vegr.T
  timeout time.Duration
  FooCalled chan bool
  FooInput struct {
   Bar chan bar.Bar
  }
  FooOutput struct {
   Ret0 chan Foo
  }
  BazCalled chan bool
  BaconCalled chan bool
  BaconInput struct {
    Arg0 chan func(Eggs) Eggs
  }
  BaconOutput struct {
    Ret0 chan func(Eggs) Eggs
  }
 }

 func newMockBar(t vegr.T, timeout time.Duration) *mockBar {
  m := &mockBar{t: t, timeout: timeout}
  m.FooCalled = make(chan bool, 100)
  m.FooInput.Bar = make(chan bar.Bar, 100)
  m.FooOutput.Ret0 = make(chan Foo, 100)
  m.BazCalled = make(chan bool, 100)
  m.BaconCalled = make(chan bool, 100)
  m.BaconInput.Arg0 = make(chan func(Eggs) Eggs, 100)
  m.BaconOutput.Ret0 = make(chan func(Eggs) Eggs, 100)
  return m
 }
 func (m *mockBar) Foo(bar bar.Bar) (ret0 Foo) {
  m.t.Helper()
  m.FooCalled <- true
  m.FooInput.Bar <- bar
  vegr.PopulateReturns(m.t, "Foo", m.timeout, m.FooOutput, &ret0)
  return ret0
 }
 func (m *mockBar) Baz() {
  m.t.Helper()
  m.BazCalled <- true
 }
 func (m *mockBar) Bacon(arg0 func(Eggs) Eggs) (ret0 func(Eggs) Eggs) {
  m.t.Helper()
  m.BaconCalled <- true
  m.BaconInput.Arg0 <- arg0
  vegr.PopulateReturns(m.t, "Bacon", m.timeout, m.BaconOutput, &ret0)
  return ret0
 }
 `))
		tt.expect(err).To(not(haveOccurred()))
		tt.expect(buf.String()).To(equal(string(expected)))
	})

	o.Spec("it changes the parameter name when it conflicts with the receiver name", func(tt testCtx) {
		decls := []ast.Decl{
			&ast.GenDecl{Specs: []ast.Spec{typeSpec(tt.expect, `
  type Foo interface {
   Foo(m string)
  }`)}},
		}

		pers.ConsistentlyReturn(tt.t, tt.finder.DependenciesOutput, nil)
		pers.Return(tt.finder.DeclsOutput, decls)
		m, err := mock.Generate(tt.finder)
		tt.expect(err).To(not(haveOccurred()))

		var buf bytes.Buffer
		m.Output("foo", "test/withoutimports", 100, &buf)

		expected, err := format.Source([]byte(`
 // Code generated by git.sr.ht/~nelsam/hel. DO NOT EDIT.
 //
 // This file contains mocks generated by hel. Do not edit this code by
 // hand unless you *really* know what you're doing. Expect any changes
 // made manually to be overwritten the next time hel regenerates this
 // file.

 package foo

 import (
  "git.sr.ht/~nelsam/hel/vegr"
  "time"
 )

 type mockFoo struct {
  t vegr.T
  timeout time.Duration
  FooCalled chan bool
  FooInput struct {
   M chan string
  }
 }

 func newMockFoo(t vegr.T, timeout time.Duration) *mockFoo {
  m := &mockFoo{t: t, timeout: timeout}
  m.FooCalled = make(chan bool, 100)
  m.FooInput.M = make(chan string, 100)
  return m
 }
 func (m *mockFoo) Foo(m_ string) {
  m.t.Helper()
  m.FooCalled <- true
  m.FooInput.M <- m_
 }
 `))
		tt.expect(err).To(not(haveOccurred()))
		tt.expect(buf.String()).To(equal(string(expected)))
	})

	o.Spec("it doesn't prepend the local package to generic types which shadow local types", func(tt testCtx) {
		decls := []ast.Decl{
			&ast.GenDecl{Specs: []ast.Spec{typeSpec(tt.expect, `
  type T struct{}
  `)}},
			&ast.GenDecl{Specs: []ast.Spec{typeSpec(tt.expect, `
  type Foo[T any] interface {
   Foo(v T)
  }`)}},
		}

		pers.ConsistentlyReturn(tt.t, tt.finder.DependenciesOutput, nil)
		pers.Return(tt.finder.DeclsOutput, decls)
		m, err := mock.Generate(tt.finder)
		tt.expect(err).To(not(haveOccurred()))

		m.PrependLocalPackage("foo")

		buf := bytes.Buffer{}
		m.Output("foo", "test/withoutimports", 100, &buf)

		expected, err := format.Source([]byte(`
 // Code generated by git.sr.ht/~nelsam/hel. DO NOT EDIT.
 //
 // This file contains mocks generated by hel. Do not edit this code by
 // hand unless you *really* know what you're doing. Expect any changes
 // made manually to be overwritten the next time hel regenerates this
 // file.

 package foo

 import (
  "git.sr.ht/~nelsam/hel/vegr"
  "time"
 )

 type mockFoo[T any] struct {
  t vegr.T
  timeout time.Duration
  FooCalled chan bool
  FooInput struct {
   V chan T
  }
 }

 func newMockFoo[T any](t vegr.T, timeout time.Duration) *mockFoo[T] {
  m := &mockFoo[T]{t: t, timeout: timeout}
  m.FooCalled = make(chan bool, 100)
  m.FooInput.V = make(chan T, 100)
  return m
 }
 func (m *mockFoo[T]) Foo(v T) {
  m.t.Helper()
  m.FooCalled <- true
  m.FooInput.V <- v
 }
 `))
		tt.expect(err).To(not(haveOccurred()))
		tt.expect(buf.String()).To(equal(string(expected)))
	})
}

func mockFor(expect expect.Expectation, decl ast.Decl) mock.Mock {
	gdecl, ok := decl.(*ast.GenDecl)
	expect(ok).To(beTrue())

	spec, ok := gdecl.Specs[0].(*ast.TypeSpec)
	expect(ok).To(beTrue())

	var params []*ast.Field
	if spec.TypeParams != nil {
		params = spec.TypeParams.List
	}
	inter, ok := spec.Type.(*ast.InterfaceType)
	expect(ok).To(beTrue())
	m, err := mock.For(spec.Name.Name, params, inter)
	expect(err).To(not(haveOccurred()))
	return m
}
