// This is free and unencumbered software released into the public
// domain.  For more information, see <http://unlicense.org> or the
// accompanying UNLICENSE file.

package pers_test

import (
	"errors"
	"fmt"
	"testing"
	"time"

	"git.sr.ht/~nelsam/hel/pkg/pers"
	"github.com/poy/onpar"
	"github.com/poy/onpar/expect"
	"github.com/poy/onpar/matchers"
)

type fakeMock struct {
	FooCalled chan struct{}
	FooInput  struct {
		Arg0 chan int
		Arg1 chan string
	}
	FooOutput struct {
		Err chan error
	}
	BarCalled chan struct{}
}

func newFakeMock() *fakeMock {
	m := &fakeMock{}
	m.FooCalled = make(chan struct{}, 100)
	m.FooInput.Arg0 = make(chan int, 100)
	m.FooInput.Arg1 = make(chan string, 100)
	m.FooOutput.Err = make(chan error, 100)
	m.BarCalled = make(chan struct{}, 100)
	return m
}

func (m *fakeMock) Foo(arg0 int, arg1 string) error {
	m.FooCalled <- struct{}{}
	m.FooInput.Arg0 <- arg0
	m.FooInput.Arg1 <- arg1
	return <-m.FooOutput.Err
}

func (m *fakeMock) Bar() {
	m.BarCalled <- struct{}{}
}

type fakeVariadicMock struct {
	FooCalled chan struct{}
	FooInput  struct {
		Args chan []*string
	}
	BarCalled chan struct{}
	BarInput  struct {
		Format chan string
		Args   chan []interface{}
	}
}

func newFakeVariadicMock() *fakeVariadicMock {
	m := &fakeVariadicMock{}
	m.FooCalled = make(chan struct{}, 100)
	m.FooInput.Args = make(chan []*string, 100)
	m.BarCalled = make(chan struct{}, 100)
	m.BarInput.Format = make(chan string, 100)
	m.BarInput.Args = make(chan []interface{}, 100)
	return m
}

func (m *fakeVariadicMock) Foo(args ...*string) {
	m.FooCalled <- struct{}{}
	m.FooInput.Args <- args
}

func (m *fakeVariadicMock) Bar(format string, args ...interface{}) {
	m.BarCalled <- struct{}{}
	m.BarInput.Format <- format
	m.BarInput.Args <- args
}

type fakeSliceMapMock struct {
	FooCalled chan struct{}
	FooInput  struct {
		Arg0 chan []interface{}
		Arg1 chan map[string]interface{}
	}
}

func newFakeSliceMapMock() *fakeSliceMapMock {
	m := &fakeSliceMapMock{}
	m.FooCalled = make(chan struct{}, 100)
	m.FooInput.Arg0 = make(chan []interface{}, 100)
	m.FooInput.Arg1 = make(chan map[string]interface{}, 100)
	return m
}

func (m *fakeSliceMapMock) Foo(arg0 []interface{}, arg1 map[string]interface{}) {
	m.FooCalled <- struct{}{}
	m.FooInput.Arg0 <- arg0
	m.FooInput.Arg1 <- arg1
}

func TestHaveMethodExecuted(t *testing.T) {
	type testCtx struct {
		t      *testing.T
		expect expectation
	}
	o := onpar.BeforeEach(onpar.New(t), func(t *testing.T) testCtx {
		return testCtx{t: t, expect: expect.New(t)}
	})
	defer o.Run()

	o.Spec("it panics if the requested method isn't found", func(tt testCtx) {
		fm := newFakeMock()

		defer func() {
			r := recover()
			tt.expect(r).To(not(beNil()))
			tt.expect(fmt.Sprintf("%v", r)).To(equal("pers: could not find method 'Invalid' on type *pers_test.fakeMock"))
		}()
		m := pers.HaveMethodExecuted("Invalid")
		m.Match(fm)
	})

	o.Spec("it drains a value off of each relevant channel", func(tt testCtx) {
		fm := newFakeMock()
		fm.FooCalled <- struct{}{}
		fm.FooInput.Arg0 <- 0
		fm.FooInput.Arg1 <- "foo"

		m := pers.HaveMethodExecuted("Foo")
		m.Match(fm)

		select {
		case <-fm.FooCalled:
			tt.t.Fatal("Expected HaveMethodExecuted to drain from the mock's FooCalled channel")
		case <-fm.FooInput.Arg0:
			tt.t.Fatal("Expected HaveMethodExecuted to drain from the mock's first FooInput channel")
		case <-fm.FooInput.Arg1:
			tt.t.Fatal("Expected HaveMethodExecuted to drain frim the mock's second FooInput channel")
		default:
		}
	})

	o.Spec("it returns a success when the method has been called", func(tt testCtx) {
		fm := newFakeMock()
		fm.FooOutput.Err <- nil
		fm.Foo(1, "foo")

		m := pers.HaveMethodExecuted("Foo")
		_, err := m.Match(fm)
		tt.expect(err).To(not(haveOccurred()))
	})

	o.Spec("it returns a failure when the method has _not_ been called", func(tt testCtx) {
		m := pers.HaveMethodExecuted("Foo")
		_, err := m.Match(newFakeMock())
		tt.expect(err).To(haveOccurred())
		tt.expect(err.Error()).To(equal("pers: expected method Foo to have been called, but it was not"))
	})

	o.Spec("it can handle methods with no input or output", func(tt testCtx) {
		fm := newFakeMock()
		fm.Bar()

		m := pers.HaveMethodExecuted("Bar")
		_, err := m.Match(fm)

		tt.expect(err).To(not(haveOccurred()))
	})

	o.Spec("it waits for a method to be called", func(tt testCtx) {
		fm := newFakeMock()
		fm.FooOutput.Err <- nil

		m := pers.HaveMethodExecuted("Foo", pers.Within(100*time.Millisecond))
		errs := make(chan error)
		go func() {
			_, err := m.Match(fm)
			errs <- err
		}()

		fm.Foo(10, "bar")
		select {
		case err := <-errs:
			tt.expect(err).To(not(haveOccurred()))
		case <-time.After(100 * time.Millisecond):
			tt.t.Fatal("Expected Match to wait for Foo to be called")
		}
	})

	type customString string

	for _, test := range []struct {
		name string
		arg0 interface{}
		arg1 interface{}
		err  error
	}{
		{
			name: "fails due to a mismatch on the first argument",
			arg0: 122,
			arg1: "this is a value",
			err:  errors.New(`pers: Foo was called with incorrect arguments: [ >123!=122<, "this is a value" ]`),
		},
		{
			name: "fails due to a mismatch on the second argument",
			arg0: 123,
			arg1: "this is a val",
			err:  errors.New(`pers: Foo was called with incorrect arguments: [ 123, "this is a val>ue!=<" ]`),
		},
		{
			name: "doesn't show Any in failing diff",
			arg0: 123,
			arg1: "this value",
			err:  errors.New(`pers: Foo was called with incorrect arguments: [ 123, "this >is a !=<value" ]`),
		},
		{
			name: "passes when arguments match",
			arg0: 123,
			arg1: "this is a value",
			err:  nil,
		},
		{
			name: "passes when Any is passed in",
			arg0: pers.Any,
			arg1: pers.Any,
			err:  nil,
		},
		{
			name: "passes when arguments are convertible",
			arg0: int32(123),
			arg1: customString("this is a value"),
		},
	} {
		test := test
		o.Spec(test.name, func(tt testCtx) {
			fm := newFakeMock()
			fm.FooOutput.Err <- nil
			fm.Foo(123, "this is a value")

			m := pers.HaveMethodExecuted("Foo", pers.WithArgs(test.arg0, test.arg1))
			_, err := m.Match(fm)
			tt.expect(err).To(equal(test.err))
		})
	}

	o.Spec("it panics if too few arguments are passed in", func(tt testCtx) {
		fm := newFakeMock()
		fm.FooOutput.Err <- nil
		fm.Foo(123, "this is a value")

		defer func() {
			r := recover()
			tt.expect(r).To(not(beNil()))
			tt.expect(fmt.Sprintf("%s", r)).To(equal("pers: incorrect number of WithArgs arguments: Foo takes exactly 2 arguments (got 1)"))
		}()
		m := pers.HaveMethodExecuted("Foo", pers.WithArgs(123))
		m.Match(fm)
	})

	o.Spec("it panics if too many arguments are passed in", func(tt testCtx) {
		fm := newFakeMock()
		fm.FooOutput.Err <- nil
		fm.Foo(123, "this is a value")

		defer func() {
			r := recover()
			tt.expect(r).To(not(beNil()))
			tt.expect(fmt.Sprintf("%s", r)).To(equal("pers: incorrect number of WithArgs arguments: Foo takes exactly 2 arguments (got 3)"))
		}()
		m := pers.HaveMethodExecuted("Foo", pers.WithArgs(123, "this is a value", pers.Any))
		m.Match(fm)
	})

	o.Spec("it panics if the types don't match", func(tt testCtx) {
		fm := newFakeMock()
		fm.FooOutput.Err <- nil
		fm.Foo(123, "this is a value")

		defer func() {
			r := recover()
			tt.expect(r).To(not(beNil()))
			tt.expect(fmt.Sprintf("%s", r)).To(equal("pers: wrong type for argument 0 in WithArgs: string is not convertible to int"))
		}()
		m := pers.HaveMethodExecuted("Foo", pers.WithArgs("123", "this is a value"))
		m.Match(fm)
	})

	o.Spec("it checks for Any in slices", func(tt testCtx) {
		fm := newFakeSliceMapMock()
		fm.Foo([]interface{}{"foo", "bar"}, nil)

		m := pers.HaveMethodExecuted("Foo", pers.WithArgs([]interface{}{pers.Any, "bar"}, nil))
		_, err := m.Match(fm)
		tt.expect(err).To(not(haveOccurred()))
	})

	o.Spec("it checks for Matcher types in slices", func(tt testCtx) {
		fm := newFakeSliceMapMock()
		fm.Foo([]interface{}{"foo", "bar"}, nil)

		m := pers.HaveMethodExecuted("Foo", pers.WithArgs([]interface{}{matchers.ContainSubstring("oo"), "bar"}, nil))
		_, err := m.Match(fm)
		tt.expect(err).To(not(haveOccurred()))
	})

	o.Spec("it checks for Any in maps", func(tt testCtx) {
		fm := newFakeSliceMapMock()
		fm.Foo(nil, map[string]interface{}{"foo": "bar"})

		m := pers.HaveMethodExecuted("Foo", pers.WithArgs(nil, map[string]interface{}{"foo": pers.Any}))
		_, err := m.Match(fm)
		tt.expect(err).To(not(haveOccurred()))
	})

	o.Spec("nil satisfies Any", func(tt testCtx) {
		fm := newFakeSliceMapMock()
		fm.Foo(nil, nil)

		m := pers.HaveMethodExecuted("Foo", pers.WithArgs(pers.Any, pers.Any))
		_, err := m.Match(fm)
		tt.expect(err).To(not(haveOccurred()))
	})

	o.Spec("it checks for Any in variadic arguments", func(tt testCtx) {
		fm := newFakeVariadicMock()
		fm.Foo(nil)

		m := pers.HaveMethodExecuted("Foo", pers.WithArgs(pers.Any))
		_, err := m.Match(fm)
		tt.expect(err).To(not(haveOccurred()))
	})

	o.Spec("it supports VariadicAny to allow any number of variadic arguments", func(tt testCtx) {
		fm := newFakeVariadicMock()
		fm.Foo(nil, nil, nil)

		m := pers.HaveMethodExecuted("Foo", pers.WithArgs(pers.VariadicAny))
		_, err := m.Match(fm)
		tt.expect(err).To(not(haveOccurred()))
	})

	o.Spec("it checks for Matcher types in maps", func(tt testCtx) {
		fm := newFakeSliceMapMock()
		fm.Foo(nil, map[string]interface{}{"foo": "bar"})

		m := pers.HaveMethodExecuted("Foo", pers.WithArgs(nil, map[string]interface{}{"foo": matchers.ContainSubstring("ar")}))
		_, err := m.Match(fm)
		tt.expect(err).To(not(haveOccurred()))
	})

	o.Spec("it handles variadic arguments", func(tt testCtx) {
		fm := newFakeVariadicMock()
		fm.FooCalled <- struct{}{}
		fm.FooInput.Args <- []*string{strPtr("foo"), strPtr("bar")}
		m := pers.HaveMethodExecuted("Foo", pers.WithArgs(strPtr("foo"), strPtr("bar")))
		_, err := m.Match(fm)
		tt.expect(err).To(not(haveOccurred()))
	})

	o.Spec("it handles empty variadic arguments", func(tt testCtx) {
		fm := newFakeVariadicMock()
		fm.FooCalled <- struct{}{}
		fm.FooInput.Args <- []*string(nil)
		m := pers.HaveMethodExecuted("Foo", pers.WithArgs())
		_, err := m.Match(fm)
		tt.expect(err).To(not(haveOccurred()))
	})

	o.Spec("it panics if too few arguments are passed in", func(tt testCtx) {
		fm := newFakeVariadicMock()
		fm.Bar("foo", 1)

		defer func() {
			r := recover()
			tt.expect(r).To(not(beNil()))
			tt.expect(fmt.Sprintf("%s", r)).To(equal("pers: incorrect number of WithArgs arguments: Bar takes at least 1 arguments (got 0)"))
		}()
		m := pers.HaveMethodExecuted("Bar", pers.WithArgs())
		m.Match(fm)
	})

	o.Spec("it panics if the variadic arguments are the wrong type", func(tt testCtx) {
		fm := newFakeVariadicMock()
		fm.FooCalled <- struct{}{}
		fm.FooInput.Args <- []*string{strPtr("foo"), strPtr("bar")}
		defer func() {
			r := recover()
			tt.expect(r).To(not(beNil()))
			tt.expect(fmt.Sprintf("%v", r)).To(equal("pers: variadic argument 1 (int) is not convertible to *string"))
		}()
		m := pers.HaveMethodExecuted("Foo", pers.WithArgs(strPtr("foo"), 1))
		m.Match(fm)
	})

	o.Spec("it handles nil values in variadic arguments", func(tt testCtx) {
		fm := newFakeVariadicMock()
		fm.FooCalled <- struct{}{}
		fm.FooInput.Args <- []*string{nil}
		m := pers.HaveMethodExecuted("Foo", pers.WithArgs(nil))
		_, err := m.Match(fm)
		tt.expect(err).To(not(haveOccurred()))
	})

	o.Spec("it can accept other matchers", func(tt testCtx) {
		fm := newFakeMock()
		fm.FooOutput.Err <- nil
		fm.Foo(123, "this is a value")

		m := pers.HaveMethodExecuted("Foo", pers.WithArgs(123, matchers.ContainSubstring("value")))
		_, err := m.Match(fm)
		tt.expect(err).To(not(haveOccurred()))
	})

	o.Spec("it panics if StoreArgs is called with too few values", func(tt testCtx) {
		fm := newFakeMock()
		fm.FooCalled <- struct{}{}
		fm.FooInput.Arg0 <- 42
		fm.FooInput.Arg1 <- "foobar"

		defer func() {
			r := recover()
			tt.expect(r).To(not(beNil()))
			tt.expect(fmt.Sprintf("%v", r)).To(equal("pers: incorrect number of StoreArgs arguments: Foo takes 2 arguments (got 1)"))
		}()
		var arg0 int
		m := pers.HaveMethodExecuted("Foo", pers.StoreArgs(&arg0))
		m.Match(fm)
	})

	o.Spec("it panics if StoreArgs is called with too many values", func(tt testCtx) {
		fm := newFakeMock()
		fm.FooCalled <- struct{}{}
		fm.FooInput.Arg0 <- 42
		fm.FooInput.Arg1 <- "foobar"

		defer func() {
			r := recover()
			tt.expect(r).To(not(beNil()))
			tt.expect(fmt.Sprintf("%v", r)).To(equal("pers: incorrect number of StoreArgs arguments: Foo takes 2 arguments (got 3)"))
		}()
		var (
			arg0 int
			arg1 string
			arg2 interface{}
		)
		m := pers.HaveMethodExecuted("Foo", pers.StoreArgs(&arg0, &arg1, &arg2))
		m.Match(fm)
	})

	o.Spec("the panic includes hints if StoreArgs is called with incorrect arguments on a variadic method", func(tt testCtx) {
		fm := newFakeVariadicMock()
		fm.FooCalled <- struct{}{}
		fm.FooInput.Args <- []*string{strPtr("foo"), strPtr("bar")}

		defer func() {
			r := recover()
			tt.expect(r).To(not(beNil()))
			tt.expect(fmt.Sprintf("%v", r)).To(equal("pers: incorrect number of StoreArgs arguments: Foo takes 1 arguments (got 2) [hint: Foo is variadic and you must provide a slice to store all variadic arguments to]"))
		}()
		var (
			arg0 *string
			arg1 *string
		)
		m := pers.HaveMethodExecuted("Foo", pers.StoreArgs(&arg0, &arg1))
		m.Match(fm)
	})

	o.Spec("it panics if StoreArgs is called with mismatched types", func(tt testCtx) {
		fm := newFakeMock()
		fm.FooCalled <- struct{}{}
		fm.FooInput.Arg0 <- 42
		fm.FooInput.Arg1 <- "foobar"

		defer func() {
			r := recover()
			tt.expect(r).To(not(beNil()))
			tt.expect(fmt.Sprintf("%v", r)).To(equal("pers: wrong type for argument 1 in StoreArgs arguments: string is not convertible to int"))
		}()
		var (
			arg0, arg1 int
		)
		m := pers.HaveMethodExecuted("Foo", pers.StoreArgs(&arg0, &arg1))
		m.Match(fm)
	})

	o.Spec("it converts convertible types for StoreArgs", func(tt testCtx) {
		fm := newFakeMock()
		fm.FooCalled <- struct{}{}
		fm.FooInput.Arg0 <- 42
		fm.FooInput.Arg1 <- "foobar"

		m := pers.HaveMethodExecuted("Foo", pers.Returning(nil))
		_, err := m.Match(fm)
		tt.expect(err).To(not(haveOccurred()))
		tt.expect(fm.FooOutput.Err).To(chain(receive(), not(haveOccurred())))
	})

	o.Spec("it returns values", func(tt testCtx) {
		fm := newFakeMock()

		m := pers.HaveMethodExecuted("Foo", pers.Within(time.Second), pers.Returning(nil))
		returned := make(chan error)
		go func() {
			defer close(returned)
			_, err := m.Match(fm)
			returned <- err
		}()

		tt.expect(fm.Foo(42, "foobar")).To(not(haveOccurred()))
		tt.expect(returned).To(chain(receive(receiveWait(time.Second)), not(haveOccurred())))
	})
}

func strPtr(v string) *string {
	return &v
}

func ExampleStoreArgs() {
	// Simulate calling a method on a mock
	fm := newFakeMock()
	fm.FooCalled <- struct{}{}
	fm.FooInput.Arg0 <- 42
	fm.FooInput.Arg1 <- "foobar"

	// Provide some addresses to store the arguments
	var (
		arg0 int
		arg1 string
	)
	m := pers.HaveMethodExecuted("Foo", pers.StoreArgs(&arg0, &arg1))
	_, err := m.Match(fm)
	fmt.Println(err)
	fmt.Println(arg0)
	fmt.Println(arg1)
	// Output:
	// <nil>
	// 42
	// foobar
}
