// This is free and unencumbered software released into the public
// domain.  For more information, see <http://unlicense.org> or the
// accompanying UNLICENSE file.

package typ

import (
	"fmt"
	"go/ast"
	"log"
	"regexp"
	"strings"
	"unicode"

	"git.sr.ht/~nelsam/hel/dep"
	"golang.org/x/tools/go/packages"
)

var (
	// errorMethod is the type of the Error method on error types.
	// It's defined here for any interface types that embed error.
	errorMethod = &ast.Field{
		Names: []*ast.Ident{{Name: "Error"}},
		Type: &ast.FuncType{
			Params: &ast.FieldList{},
			Results: &ast.FieldList{
				List: []*ast.Field{{Type: &ast.Ident{Name: "string"}}},
			},
		},
	}
)

// A GoDir is a type that represents a directory of Go files.
type GoDir interface {
	Path() (path string)
	Package() *packages.Package
	Import(path string) (*packages.Package, error)
}

// Dependency is a type for tracking dependencies of local logic.
type Dependency interface {
	Name() string
	TypeName() string
	PkgPath() string
	PkgName() string
	ImportAliased() bool
	Type() ast.Expr
	Params() []*ast.Field
	AvoidCollision(name string) dep.Dependency
	FromPkg(name, path string) dep.Dependency
}

// A Dir is a type that represents a directory containing Go
// packages.
type Dir struct {
	dir          string
	pkg          string
	dependencies map[string][]Dependency
	decls        []ast.Decl
}

// Clone returns a copy of d.
func (d *Dir) Clone() *Dir {
	copy := *d
	copy.dependencies = make(map[string][]Dependency)
	for k, v := range d.dependencies {
		copy.dependencies[k] = append([]Dependency(nil), v...)
	}
	copy.decls = append([]ast.Decl(nil), d.decls...)
	return &copy
}

// Dir returns the directory path that d represents.
func (d *Dir) Dir() string {
	return d.dir
}

// Len returns the number of types that will be returned by
// d.ExportedDecls().
func (d *Dir) Len() int {
	return len(d.decls)
}

// Package returns the name of d's importable package.
func (d *Dir) Package() string {
	return d.pkg
}

// Decls returns all ast.Decl (declarations) found by d. Interface types with
// anonymous interface types will be flattened, for ease of mocking by other
// logic.
//
// This includes unexported declaratiions.
func (d *Dir) Decls() []ast.Decl {
	return d.decls
}

// Dependencies returns all interface types that typ depends on for
// method parameters or results.
func (d *Dir) Dependencies(name string) []Dependency {
	added := map[string]struct{}{
		name: {},
	}
	return d.recurseDeps(added, name)
}

func (d *Dir) recurseDeps(added map[string]struct{}, name string) []Dependency {
	var deps []Dependency
	for _, dep := range d.dependencies[name] {
		if _, ok := added[dep.Name()]; ok {
			continue
		}
		added[dep.Name()] = struct{}{}
		deps = append(deps, dep)
		deps = append(deps, d.recurseDeps(added, dep.Name())...)
	}
	return deps
}

func newDir(dirs Dirs, pkg *packages.Package, dir GoDir, pkgName, pkgPath string, unexported bool) *Dir {
	d := &Dir{
		pkg:          pkgName,
		dir:          pkgPath,
		dependencies: make(map[string][]Dependency),
	}
	newDecls, _, depMap := dirs.loadPkgDecls(pkg, dir, unexported)
	if d.pkg == "" {
		d.pkg = pkg.Name
	}
	for name, deps := range depMap {
		d.dependencies[name] = append(d.dependencies[name], deps...)
	}
	d.decls = append(d.decls, newDecls...)
	return d
}

func matchingSpecs(specs []ast.Spec, matchers ...*regexp.Regexp) []ast.Spec {
	var matching []ast.Spec
	for _, s := range specs {
		switch spec := s.(type) {
		case *ast.TypeSpec:
			for _, matcher := range matchers {
				if !matcher.MatchString(spec.Name.Name) {
					continue
				}
				matching = append(matching, spec)
				break
			}
		default:
			continue
		}
	}
	return matching
}

// Filter returns a copy of d with all types filtered. All types in the returned
// Dir will match at least one of the passed in matchers.
func (d *Dir) Filter(matchers ...*regexp.Regexp) *Dir {
	copy := d.Clone()
	copy.decls = nil
	for _, decl := range d.Decls() {
		switch dt := decl.(type) {
		case *ast.GenDecl:
			newDT := *dt
			newDT.Specs = matchingSpecs(dt.Specs, matchers...)
			if len(newDT.Specs) > 0 {
				copy.decls = append(copy.decls, &newDT)
			}
		case *ast.FuncDecl:
			for _, matcher := range matchers {
				if !matcher.MatchString(dt.Name.Name) {
					continue
				}
				copy.decls = append(copy.decls, decl)
				break
			}
		}
	}
	return copy
}

// Dirs is a slice of Dir values, to provide sugar for running some
// methods against multiple Dir values.
type Dirs struct {
	dirs []*Dir

	importCache map[string]*Dir
}

// Load loads a Dirs value for goDirs.
func Load(unexported bool, goDirs ...GoDir) Dirs {
	d := Dirs{
		importCache: make(map[string]*Dir),
	}
	for _, dir := range goDirs {
		d.dirs = append(d.dirs, newDir(d, dir.Package(), dir, dir.Package().Name, dir.Path(), unexported))
	}
	return d
}

// Slice returns the Dir values contained in d in slice form.
func (d Dirs) Slice() []*Dir {
	return d.dirs
}

// Filter calls Dir.Filter for each Dir in d.
func (d Dirs) Filter(patterns ...string) Dirs {
	if len(patterns) == 0 {
		return d
	}
	matchers := make([]*regexp.Regexp, 0, len(patterns))
	for _, pattern := range patterns {
		matchers = append(matchers, regexp.MustCompile("^"+pattern+"$"))
	}
	newDirs := Dirs{
		importCache: d.importCache,
	}
	for _, dir := range d.dirs {
		copy := dir.Filter(matchers...)
		if copy.Len() > 0 {
			newDirs.dirs = append(newDirs.dirs, copy)
		}
	}
	return newDirs
}

func isAny[T, U any](l []T, translate func(T) U, check func(U) bool) bool {
	for _, v := range l {
		if check(translate(v)) {
			return true
		}
	}
	return false
}

type namedDecl struct {
	name string
	decl ast.Decl
}

type depVisitor struct {
	available  []*ast.TypeSpec
	imports    []*ast.ImportSpec
	dir        GoDir
	d          Dirs
	unexported bool

	deps map[string]Dependency
}

func (v *depVisitor) Visit(n ast.Node) ast.Visitor {
	switch n := n.(type) {
	case *ast.FuncDecl:
		if !v.unexported && !unicode.IsUpper([]rune(n.Name.Name)[0]) {
			return v
		}
		addSpecs(v.deps, v.d.loadFieldDependencies(n.Type.Params, v.available, v.imports, v.dir)...)
		return v
	case *ast.TypeSpec:
		switch t := n.Type.(type) {
		case *ast.InterfaceType:
			if t.Methods == nil {
				return v
			}
			for _, meth := range t.Methods.List {
				f, ok := meth.Type.(*ast.FuncType)
				if !ok {
					panic(fmt.Errorf("typ: unexpected non-function type in interface %v", t))
				}
				addSpecs(v.deps, v.d.loadFieldDependencies(f.Params, v.available, v.imports, v.dir)...)
				addSpecs(v.deps, v.d.loadFieldDependencies(f.Results, v.available, v.imports, v.dir)...)
			}
			return v
		case *ast.StructType:
			if !v.unexported {
				allFields := t.Fields.List
				t.Fields.List = nil
				for _, f := range allFields {
					firstChar := func(n *ast.Ident) rune { return []rune(n.Name)[0] }
					if !isAny(f.Names, firstChar, unicode.IsUpper) {
						continue
					}
					t.Fields.List = append(t.Fields.List, f)
				}
			}
			addSpecs(v.deps, v.d.loadFieldDependencies(t.Fields, v.available, v.imports, v.dir)...)
			return v
		default:
			return v
		}
	default:
		return v
	}
}

// dependencies returns all types that decl depends on in a testing context.
// This includes parameter/return types for methods on interfaces, parameter
// types for functions, and field types for structs, to name a few.
func (d Dirs) dependencies(decl ast.Decl, available []*ast.TypeSpec, withImports []*ast.ImportSpec, dir GoDir, unexported bool) []Dependency {
	v := &depVisitor{
		available:  available,
		imports:    withImports,
		dir:        dir,
		d:          d,
		unexported: unexported,
		deps:       make(map[string]Dependency),
	}
	ast.Walk(v, decl)
	depSlice := make([]Dependency, 0, len(v.deps))
	for _, dep := range v.deps {
		depSlice = append(depSlice, dep)
	}
	return depSlice
}

func addSpecs(set map[string]Dependency, values ...Dependency) {
	for _, value := range values {
		set[value.Name()] = value
	}
}

func names(specs []*ast.TypeSpec) (names []string) {
	for _, spec := range specs {
		names = append(names, spec.Name.String())
	}
	return names
}

func (d Dirs) loadFieldDependencies(fields *ast.FieldList, available []*ast.TypeSpec, withImports []*ast.ImportSpec, dir GoDir) (deps []Dependency) {
	if fields == nil {
		return nil
	}
	for _, field := range fields.List {
		deps = append(deps, d.loadTypeDependencies(field.Type, available, withImports, dir)...)
	}
	return deps
}

func (d Dirs) loadTypeDependencies(typ ast.Expr, available []*ast.TypeSpec, withImports []*ast.ImportSpec, dir GoDir) []Dependency {
	switch src := typ.(type) {
	case *ast.ArrayType:
		return d.loadTypeDependencies(src.Elt, available, withImports, dir)
	case *ast.MapType:
		deps := d.loadTypeDependencies(src.Key, available, withImports, dir)
		deps = append(deps, d.loadTypeDependencies(src.Value, available, withImports, dir)...)
		return deps
	case *ast.StarExpr:
		return d.loadTypeDependencies(src.X, available, withImports, dir)
	case *ast.Ellipsis:
		return d.loadTypeDependencies(src.Elt, available, withImports, dir)
	case *ast.ChanType:
		return d.loadTypeDependencies(src.Value, available, withImports, dir)
	case *ast.Ident:
		spec := findSpec(src, available)
		if spec == nil {
			return nil
		}
		if _, ok := spec.Type.(*ast.InterfaceType); !ok {
			return nil
		}
		var opts []dep.Opt
		if spec.TypeParams != nil {
			opts = append(opts, dep.WithTypeParams(spec.TypeParams.List...))
		}
		return []Dependency{dep.New(spec.Name, spec.Type, opts...)}
	case *ast.SelectorExpr:
		selectorName := src.X.(*ast.Ident).String()
		for _, imp := range withImports {
			if imp.Name != nil && imp.Name.Name != selectorName {
				continue
			}
			importPath := strings.Trim(imp.Path.Value, `"`)
			impDir, ok := d.importCache[importPath]
			if !ok {
				pkg, err := dir.Import(importPath)
				if err != nil {
					continue
				}
				if imp.Name == nil && pkg.Name != selectorName {
					// The import has been cached, but we don't need to process
					// the types just yet.
					continue
				}
				impDir = newDir(d, pkg, dir, "", "", false)
				d.importCache[importPath] = impDir
			}
			opts := []dep.Opt{dep.FromPkg(impDir.pkg, importPath)}
			importName := impDir.pkg
			if imp.Name != nil {
				importName = imp.Name.Name
				opts = append(opts, dep.WithImportAlias(imp.Name.Name))
			}
			for _, decl := range impDir.Decls() {
				switch dt := decl.(type) {
				case *ast.GenDecl:
					for _, spec := range dt.Specs {
						switch st := spec.(type) {
						case *ast.TypeSpec:
							if st.Name.Name != src.Sel.String() {
								continue
							}
							if st.TypeParams != nil {
								opts = append(opts, dep.WithTypeParams(st.TypeParams.List...))
							}
							dependencies := []Dependency{dep.New(st.Name, st.Type, opts...)}
							switch st.Type.(type) {
							case *ast.InterfaceType:
								// Only interface types need to recurse to sub-dependencies,
								// because concrete sub-dependencies won't be called in a
								// way that allows mocks to be used. Non-interface types are
								// really only added to the dependency list so we know what
								// to add to the import clause.
								subDeps := impDir.Dependencies(st.Name.Name)
								for i, d := range subDeps {
									if d.PkgPath() != "" {
										continue
									}
									subDeps[i] = d.FromPkg(importName, importPath)
								}
								return append(dependencies, subDeps...)
							default:
								return dependencies
							}
						default:
							continue
						}
					}
				case *ast.FuncDecl:
					if dt.Name.Name != src.Sel.String() {
						continue
					}
					if dt.Type.TypeParams != nil {
						opts = append(opts, dep.WithTypeParams(dt.Type.TypeParams.List...))
					}
					return []Dependency{dep.New(dt.Name, dt.Type, opts...)}
				}
			}
		}
		return nil
	case *ast.FuncType:
		return d.loadFieldDependencies(src.Params, available, withImports, dir)
	default:
		return nil
	}
}

func (d Dirs) loadPkgDecls(pkg *packages.Package, dir GoDir, unexported bool) (decls []ast.Decl, imports map[string][]*ast.ImportSpec, depMap map[string][]Dependency) {
	// TODO: if/when performance starts getting hairy, this is a good place to
	// skip unexported symbols. Unexported interface types should probably still
	// be processed since they _technically_ can be used by exported logic, but
	// unexported functions and the like could be skipped.
	depMap = make(map[string][]Dependency)
	imports = make(map[string][]*ast.ImportSpec)
	var types []*ast.TypeSpec
	defer func() {
		for _, decl := range decls {
			switch decl := decl.(type) {
			case *ast.GenDecl:
				for _, spec := range decl.Specs {
					switch spec := spec.(type) {
					case *ast.TypeSpec:
						depMap[spec.Name.Name] = append(depMap[spec.Name.Name], d.dependencies(decl, types, imports[spec.Name.Name], dir, unexported)...)
					default:
						continue
					}
				}
			case *ast.FuncDecl:
				name := decl.Name.Name
				if decl.Recv != nil && len(decl.Recv.List) > 0 {
					ident, ok := decl.Recv.List[0].Type.(*ast.Ident)
					if ok {
						name = ident.Name
					}
				}
				depMap[name] = append(depMap[name], d.dependencies(decl, types, imports[decl.Name.Name], dir, unexported)...)
				continue
			default:
				continue
			}
		}
	}()
	for _, f := range pkg.Syntax {
		fileImports := f.Imports
		fileDecls := loadFileDecls(f)
		var fileTypes []*ast.TypeSpec
		for _, decl := range fileDecls {
			switch decl := decl.(type) {
			case *ast.GenDecl:
				for _, spec := range decl.Specs {
					typ, ok := spec.(*ast.TypeSpec)
					if !ok {
						continue
					}
					imports[typ.Name.Name] = fileImports
					fileTypes = append(fileTypes, typ)
				}
			case *ast.FuncDecl:
				imports[decl.Name.Name] = fileImports
			default:
				continue
			}
		}

		// flattenAnon needs to be called for each file, but the
		// withSpecs parameter needs *all* specs, from *all* files.
		// So we defer the flatten call until all files are processed.
		defer func() {
			d.flattenAnon(fileTypes, types, fileImports, imports, dir, unexported)
		}()

		decls = append(decls, fileDecls...)
		types = append(types, fileTypes...)
	}
	return decls, imports, depMap
}

func loadFileDecls(f *ast.File) (decls []ast.Decl) {
	for _, d := range f.Decls {
		switch decl := d.(type) {
		case *ast.GenDecl:
			// This consists of both concrete and interface types.
			decls = append(decls, decl)
		case *ast.FuncDecl:
			decls = append(decls, decl)
		default:
			continue
		}
	}
	return decls
}

func (d Dirs) flattenAnon(specs, withSpecs []*ast.TypeSpec, fileImports []*ast.ImportSpec, declImports map[string][]*ast.ImportSpec, dir GoDir, unexported bool) {
	for _, spec := range specs {
		d.flatten(spec, withSpecs, fileImports, declImports, dir, unexported)
	}
}

func (d Dirs) flatten(spec *ast.TypeSpec, withSpecs []*ast.TypeSpec, fileImports []*ast.ImportSpec, declImports map[string][]*ast.ImportSpec, dir GoDir, unexported bool) {
	switch t := spec.Type.(type) {
	case *ast.InterfaceType:
		if t.Methods == nil {
			return
		}
		methods := make([]*ast.Field, 0, len(t.Methods.List))
		for _, method := range t.Methods.List {
			switch src := method.Type.(type) {
			case *ast.FuncType:
				methods = append(methods, method)
			case *ast.Ident:
				methods = append(methods, d.findAnonMethods(src, withSpecs, declImports[src.Name], declImports, dir, unexported)...)
			case *ast.SelectorExpr:
				importedTypes, imports, _ := d.findImportedTypes(src.X.(*ast.Ident), fileImports, dir)
				methods = append(methods, d.findAnonMethods(src.Sel, importedTypes, imports[src.Sel.Name], imports, dir, unexported)...)
			}
		}
		t.Methods.List = methods
	case *ast.Ident:
		// type alias (in the same package)
		realSpec := findSpec(t, withSpecs)
		if realSpec == nil {
			return
		}
		spec.Type = realSpec.Type
		d.flatten(spec, withSpecs, declImports[t.Name], declImports, dir, unexported)
	case *ast.SelectorExpr:
		imported, imports, _ := d.findImportedTypes(t.X.(*ast.Ident), fileImports, dir)
		realSpec := findSpec(t.Sel, imported)
		if realSpec == nil {
			return
		}
		spec.Type = realSpec.Type
		d.flatten(spec, imported, imports[realSpec.Name.Name], imports, dir, unexported)
	}
}

func (d Dirs) findImportedTypes(name *ast.Ident, withImports []*ast.ImportSpec, dir GoDir) ([]*ast.TypeSpec, map[string][]*ast.ImportSpec, map[string][]Dependency) {
	importName := name.String()
	for _, imp := range withImports {
		path := strings.Trim(imp.Path.Value, `"`)
		pkg, err := dir.Import(path)
		if err != nil {
			log.Printf("Error loading import '%s': %s", path, err)
			continue
		}
		name := pkg.Name
		if imp.Name != nil {
			name = imp.Name.String()
		}
		if name != importName {
			continue
		}
		decls, imports, deps := d.loadPkgDecls(pkg, dir, false)
		var types []*ast.TypeSpec
		for _, d := range decls {
			switch decl := d.(type) {
			case *ast.GenDecl:
				for _, s := range decl.Specs {
					typ, ok := s.(*ast.TypeSpec)
					if !ok {
						continue
					}
					types = append(types, typ)
				}
			case *ast.FuncDecl:
				// TODO: recursive function/type dependencies?
			default:
				continue
			}
		}
		addSelector(types, importName)
		return types, imports, deps
	}
	return nil, nil, nil
}

func addSelector(typs []*ast.TypeSpec, selector string) {
	for _, typ := range typs {
		inter, ok := typ.Type.(*ast.InterfaceType)
		if !ok {
			continue
		}
		for _, meth := range inter.Methods.List {
			addFuncSelectors(meth.Type.(*ast.FuncType), selector)
		}
	}
}

func addFuncSelectors(method *ast.FuncType, selector string) {
	if method.Params != nil {
		addFieldSelectors(method.Params.List, selector)
	}
	if method.Results != nil {
		addFieldSelectors(method.Results.List, selector)
	}
}

func addFieldSelectors(fields []*ast.Field, selector string) {
	for idx, field := range fields {
		fields[idx] = addFieldSelector(field, selector)
	}
}

func addFieldSelector(field *ast.Field, selector string) *ast.Field {
	switch src := field.Type.(type) {
	case *ast.Ident:
		if !unicode.IsUpper(rune(src.String()[0])) {
			return field
		}
		return &ast.Field{
			Type: &ast.SelectorExpr{
				X:   &ast.Ident{Name: selector},
				Sel: src,
			},
		}
	case *ast.FuncType:
		addFuncSelectors(src, selector)
	}
	return field
}

func findSpec(ident *ast.Ident, inSpecs []*ast.TypeSpec) *ast.TypeSpec {
	for _, spec := range inSpecs {
		if spec.Name.String() == ident.Name {
			return spec
		}
	}
	return nil
}

func (d Dirs) findAnonMethods(ident *ast.Ident, withSpecs []*ast.TypeSpec, fileImports []*ast.ImportSpec, declImports map[string][]*ast.ImportSpec, dir GoDir, unexported bool) []*ast.Field {
	spec := findSpec(ident, withSpecs)
	if spec == nil {
		if ident.Name != "error" {
			// TODO: do something nicer with this error.
			panic(fmt.Errorf("Can't find anonymous type %s", ident.Name))
		}
		return []*ast.Field{errorMethod}
	}
	d.flatten(spec, withSpecs, fileImports, declImports, dir, unexported)
	anon := spec.Type.(*ast.InterfaceType)
	return anon.Methods.List
}
