package vegr_test

import (
	"fmt"
	"sync"
	"testing"
	"time"

	"git.sr.ht/~nelsam/hel/vegr"
	"github.com/poy/onpar"
	"github.com/poy/onpar/expect"
)

type mockT struct {
	mu            sync.Mutex
	helperCalled  bool
	fataled       bool
	failed        bool
	fatalMsg      string
	fatalPanicMsg string
}

func (t *mockT) Helper() {
	t.mu.Lock()
	defer t.mu.Unlock()
	t.helperCalled = true
}

func (t *mockT) Failed() bool {
	t.mu.Lock()
	defer t.mu.Unlock()
	return t.failed
}

func (t *mockT) Fatalf(format string, args ...interface{}) {
	t.mu.Lock()
	defer t.mu.Unlock()
	t.fataled = true
	t.fatalMsg = fmt.Sprintf(format, args...)
	if t.fatalPanicMsg != "" {
		panic(t.fatalPanicMsg)
	}
}

func TestPopulateReturns(t *testing.T) {
	strPtr := func() *string {
		var v string
		return &v
	}

	type testCtx struct {
		expect expectation
		mT     *mockT
	}

	o := onpar.BeforeEach(onpar.New(t), func(t *testing.T) testCtx {
		return testCtx{expect: expect.New(t), mT: &mockT{}}
	})
	defer o.Run()

	for _, tt := range []struct {
		name       string
		prepMockT  func(*mockT)
		methodName string
		mock       func() interface{}
		addrs      []interface{}
		panicMsg   string
	}{
		{name: "for string types", mock: func() interface{} { return "foo" }, panicMsg: "hel: PopulateReturns was called with incorrect mock type (string): sel: type string is neither a struct of channels nor a channel"},
		{
			name: "for non-channel fields in struct types",
			mock: func() interface{} {
				return struct {
					Foo chan struct{}
					Bar int
				}{
					Foo: make(chan struct{}),
					Bar: 10,
				}
			},
			panicMsg: "hel: PopulateReturns was called with incorrect mock type (struct { Foo chan struct {}; Bar int }): sel: field 1 of type struct { Foo chan struct {}; Bar int } cannot be used as a mock field: sel: type int is neither a struct of channels nor a channel",
		},
		{
			name:     "for non-pointer arguments",
			mock:     func() interface{} { return make(chan string) },
			addrs:    []interface{}{"foo"},
			panicMsg: "hel: PopulateReturns was called with non-pointer type (string)",
		},
		{
			name:     "for nil pointer arguments",
			mock:     func() interface{} { return make(chan string) },
			addrs:    []interface{}{(*string)(nil)},
			panicMsg: "hel: PopulateReturns was called with nil pointer of type (*string)",
		},
		{
			name: "for closed channels",
			mock: func() interface{} {
				ch := make(chan string)
				close(ch)
				return ch
			},
			addrs:    []interface{}{strPtr()},
			panicMsg: "hel: PopulateReturns called on closed mock (type chan string)",
		},
		{
			name: "with a panic from t.Fatalf",
			prepMockT: func(mT *mockT) {
				mT.fatalPanicMsg = "Fail in goroutine after test has completed"
			},
			methodName: "Foo",
			mock: func() interface{} {
				return struct {
					Foo chan string
				}{
					Foo: make(chan string, 10),
				}
			},
			addrs:    []interface{}{strPtr()},
			panicMsg: "hel: panic calling t.Fatalf from mock method Foo: Fail in goroutine after test has completed",
		},
	} {
		tt := tt
		if tt.prepMockT == nil {
			tt.prepMockT = func(*mockT) {}
		}
		o.Group(tt.name, func() {
			o.Spec("it panics", func(test testCtx) {
				tt.prepMockT(test.mT)
				defer func() {
					r := recover()
					test.expect(r).To(not(beNil()))
					test.expect(fmt.Sprintf("%v", r)).To(equal(tt.panicMsg))
				}()
				vegr.PopulateReturns(test.mT, tt.methodName, time.Millisecond, tt.mock(), tt.addrs...)
			})

			o.Spec("it does not panic if the t has already failed", func(test testCtx) {
				tt.prepMockT(test.mT)
				defer func() {
					r := recover()
					test.expect(r).To(beNil())
				}()
				test.mT.failed = true
				vegr.PopulateReturns(test.mT, tt.methodName, time.Millisecond, tt.mock(), tt.addrs...)
			})
		})
	}

	o.Spec("it calls t.Helper() immediately", func(test testCtx) {
		defer func() {
			r := recover()
			test.expect(r).To(not(beNil()))
			test.expect(test.mT.helperCalled).To(beTrue())
		}()
		vegr.PopulateReturns(test.mT, "", time.Millisecond, "foo")
	})

	o.Spec("it fails the T if the return times out", func(test testCtx) {
		mock := make(chan string)
		var v string
		vegr.PopulateReturns(test.mT, "Foo", time.Millisecond, mock, &v)
		test.expect(test.mT.fataled).To(beTrue())
		test.expect(test.mT.fatalMsg).To(equal("hel: mock method Foo timed out after 1ms waiting for return on type (chan string)"))
	})

	o.Spec("it does not fail the T if the test has already failed", func(test testCtx) {
		mock := make(chan string)
		var v string
		done := make(chan struct{})
		go func() {
			defer close(done)
			vegr.PopulateReturns(test.mT, "Foo", 10*time.Millisecond, mock, &v)
		}()

		// This is intended to assert that if the test fails while
		// PopulateReturns is waiting on return values, PopulateReturns will not
		// call t.Fatalf (because the panic in t.Fatalf ends up hiding the
		// _real_ test failure). To ensure that this test will fail when
		// PopulateReturns only checks t.Failed _before_ waiting on the timeout
		// channel, we run PopulateReturns concurrently and update t.failed
		// after the PopulateReturns goroutine has started.
		time.Sleep(time.Millisecond)
		test.mT.mu.Lock()
		test.mT.failed = true
		test.mT.mu.Unlock()

		<-done
		test.expect(test.mT.fataled).To(not(beTrue()))
	})

	o.Spec("it sets the returned value to the address provided", func(test testCtx) {
		mock := make(chan string, 1)
		mock <- "foobar"
		var v string
		vegr.PopulateReturns(test.mT, "", time.Millisecond, mock, &v)
		test.expect(v).To(equal("foobar"))
		test.expect(test.mT.fataled).To(not(beTrue()))
	})

	o.Spec("it supports structs of channels", func(test testCtx) {
		mock := struct {
			Foo chan string
			Bar chan int
		}{
			Foo: make(chan string, 1),
			Bar: make(chan int, 1),
		}
		mock.Foo <- "foo"
		mock.Bar <- 10
		var (
			foo string
			bar int
		)
		vegr.PopulateReturns(test.mT, "", time.Millisecond, mock, &foo, &bar)
		test.expect(foo).To(equal("foo"))
		test.expect(bar).To(equal(10))
		test.expect(test.mT.fataled).To(not(beTrue()))
	})

	o.Spec("it does not load the same field twice", func(test testCtx) {
		mock := struct {
			Foo chan string
			Bar chan int
		}{
			Foo: make(chan string),
			Bar: make(chan int, 2),
		}
		mock.Bar <- 10
		mock.Bar <- 11
		var (
			foo string
			bar int
		)
		vegr.PopulateReturns(test.mT, "", time.Millisecond, mock, &foo, &bar)
		test.expect(bar).To(equal(10))
	})
}
